/*
 * Gnophone: A client for the Asterisk PBX
 *
 * Copyright (C) 2000, Linux Support Services, Inc.
 *
 * Written by Mark Spencer
 *
 * Linux/UNIX version distributed under the terms of
 * the GNU General Public License
 *
 * audio-control.c: Generic Audio/Telephony Interface
 *
 */


#include <gtk/gtk.h>
#include <stdio.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>

#include "gnophone.h"
#include "audio.h"

/* Target a peak at 10000 for normal speech */
#define TARGET 10000.0

static char INSTRUCTIONS[] = 
"Read this sentence over and over in a normal voice as though\n"
"you were talking to someone over the internet.  We want the level\n"
"of your sound to be at half the level on the bar.\n";

GtkWidget *audiobox = NULL;

GtkWidget *audiobox;
GtkWidget *configure;
GtkWidget *abutton;

GtkWidget *hscale, *hscale2, *hscale3;
GtkObject *adj, *adj2, *adj3;

GtkWidget *level;

GtkWidget *instruct;

GtkWidget *autoadj;

int mixer_fd = 0;
int micvol, mainvol, recvol;

static void update_mixer_mainvol(GtkAdjustment *);
static void update_mixer_micvol(GtkAdjustment *);
static void update_mixer_recvol(GtkAdjustment *);
static void open_mixerfd(void);

void (*other_loudness)(int level);


static int audiodestroy(void)
{
	gtk_grab_remove(audiobox);
	other_loudness = NULL;
	audiobox = NULL;
	return TRUE;
}

static void audio_cancel(void)
{
	gtk_widget_destroy(audiobox);
}

static void audio_okay(void)
{
	gtk_widget_destroy(audiobox);
}

static int showmenu(GtkMenu *m, GdkEvent *event)
{
    if (event->type == GDK_BUTTON_PRESS) {
		GdkEventButton *bevent = (GdkEventButton *) event; 
		gtk_menu_popup (m, NULL, NULL, NULL, NULL,
                              bevent->button, bevent->time);
		return TRUE;
    }
    return FALSE;
}

static void select_driver(GtkWidget *w, struct audio_channel *c)
{
	fprintf(stderr, "User selected channel %s\n", c->name);
	audio_select(c);
	if (audioc)
		gtk_label_set_text(GTK_LABEL(GTK_BIN(abutton)->child), audioc->name);
	else
		gtk_label_set_text(GTK_LABEL(GTK_BIN(abutton)->child), "No device selected");
	if (audioc && audioc->configure)
		gtk_widget_set_sensitive(configure, TRUE);
	else
		gtk_widget_set_sensitive(configure, FALSE);
}

#define DELTA 0.01

#define PASSES 10

static float historicmax = 0.0;
static int goodsamps = 0;
static int pass = 0;

static void update_prog_bar(int lev)
{
	int max = lev;
	float amt;
	if (max > historicmax)
		historicmax = max;
	else
		historicmax = (1 - DELTA) * historicmax + DELTA * (float)max;
	amt = historicmax / (TARGET * 2.0);
	if (amt > 1.0)
		amt = 1.0;
	printf("historicmax = %d, %f, %f, %d\n", max, historicmax, amt, goodsamps);
	if (!(pass++ % PASSES)) {
		if (amt < 0.5) {
			goodsamps = 0;
			if ((recvol >> 8) < 100)
				gtk_adjustment_set_value(GTK_ADJUSTMENT(adj3), (recvol >> 8) + 1);
			else
				gtk_adjustment_set_value(GTK_ADJUSTMENT(adj2), (micvol >> 8) + 1);
		} else if (amt > 0.7) {
			goodsamps = 0;
			if ((micvol >> 8) > 0)
				gtk_adjustment_set_value(GTK_ADJUSTMENT(adj2), (micvol >> 8) - 1);
			else
				gtk_adjustment_set_value(GTK_ADJUSTMENT(adj3), (recvol >> 8) - 1);
#if 0
			/* Reset historic max */
			historicmax *= 0;
#endif			
		} else 
			goodsamps++;
		if (goodsamps > 10)
			gtk_button_clicked(GTK_BUTTON(autoadj));
	}
	gtk_progress_bar_update(GTK_PROGRESS_BAR(level), amt);
	return;
}

static void audio_adjust(void)
{
	goodsamps = 0;
	pass = 0;
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(autoadj))) {
		gtk_widget_show(instruct);
		gtk_widget_show(level);
		other_loudness = update_prog_bar;
	} else {
		gtk_widget_hide(instruct);
		gtk_widget_hide(level);
		gtk_container_queue_resize(GTK_CONTAINER(audiobox));
		other_loudness = NULL;
	}
		
}

static int build_option_screen(void)
{
	GtkWidget *vbox;
	GtkWidget *frame;
	GtkWidget *button;
	GtkWidget *outerbox;
	GtkWidget *buttonbox;
//	GtkWidget *hbox;
	GtkWidget *menu;
	GtkWidget *mi;
	GtkWidget *labels;

	struct audio_channel *ac;
	
	/* Build a vbox */
	vbox = gtk_vbox_new(FALSE, 5);
	
	/* Build a frame for the adjustments */
	frame = gtk_frame_new("Audio Device");
	
	/* Make a button for our menu */
	if (audioc) 
		abutton = gtk_button_new_with_label(audioc->name);
	else {
		if (!aclist) {
			abutton = gtk_button_new_with_label("No Audio Devices Detected");
			gtk_widget_set_sensitive(abutton, FALSE);
		} else
			abutton = gtk_button_new_with_label("No Device Selected");
	}
	
	/* Build the menu of audio devices */
	menu = gtk_menu_new();
	ac = aclist;
	while(ac) {
		mi = gtk_menu_item_new_with_label(ac->name);
		gtk_widget_show(mi);
		gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(select_driver), ac);
		gtk_menu_append(GTK_MENU(menu), mi);
		ac = ac->next;
	}

	gtk_signal_connect_object(GTK_OBJECT(abutton), "event", GTK_SIGNAL_FUNC (showmenu), GTK_OBJECT(menu));	
	
	/* Put the button in the vbox */
	gtk_box_pack_start(GTK_BOX(vbox), abutton, FALSE, FALSE, 5);
	
	/* Make a smaller hbox for the configure button
	hbox = gtk_hbox_new(FALSE, 5);

	configure = gtk_button_new_with_label("Configure...");
	gtk_widget_set_usize(configure, 200, -1);
	if (!audioc || !audioc->configure)
		gtk_widget_set_sensitive(configure, FALSE);

	gtk_box_pack_end(GTK_BOX(hbox), configure, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

*/
	/* Add a border to the vbox and put it in the frame */
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
	gtk_container_add(GTK_CONTAINER(frame), vbox);

	/* An outer box for the frame and the buttons */
	outerbox = gtk_vbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(outerbox), frame, FALSE, FALSE, 5);

	/* Build a frame for the adjustments */
	frame = gtk_frame_new("Audio Mixer");
	vbox = gtk_vbox_new(FALSE, 5);

	adj = gtk_adjustment_new (0,0,110,10,10,10);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed", GTK_SIGNAL_FUNC(update_mixer_mainvol), NULL);

	labels = gtk_label_new("Main volume:");
	gtk_widget_show(labels);

	gtk_box_pack_start(GTK_BOX(vbox), labels, TRUE, TRUE, 0);

	hscale = gtk_hscale_new(GTK_ADJUSTMENT (adj));
	gtk_widget_show(hscale);

	gtk_box_pack_start(GTK_BOX(vbox), hscale, TRUE, TRUE, 0);

	labels = gtk_label_new("Mic. volume:");
	gtk_widget_show(labels);

	gtk_box_pack_start(GTK_BOX(vbox), labels, TRUE, TRUE, 0);

	adj2 = gtk_adjustment_new(0,0,110,10,10,10);
	gtk_signal_connect(GTK_OBJECT(adj2), "value_changed", GTK_SIGNAL_FUNC(update_mixer_micvol), NULL);

	hscale2 = gtk_hscale_new(GTK_ADJUSTMENT(adj2));
	gtk_widget_show(hscale2);

	gtk_box_pack_start(GTK_BOX(vbox), hscale2, TRUE, TRUE, 0);

	labels = gtk_label_new("Rec. volume:");
	gtk_widget_show(labels);

	gtk_box_pack_start(GTK_BOX(vbox), labels, TRUE, TRUE, 0);

	adj3 = gtk_adjustment_new(0,0,110,10,10,10);
	gtk_signal_connect(GTK_OBJECT(adj3), "value_changed", GTK_SIGNAL_FUNC(update_mixer_recvol), NULL);

	hscale3 = gtk_hscale_new(GTK_ADJUSTMENT(adj3));
	gtk_widget_show(hscale3);

	gtk_box_pack_start(GTK_BOX(vbox), hscale3, TRUE, TRUE, 0);
	/* Add a border to the vbox and put it in the frame */
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
	gtk_container_add(GTK_CONTAINER(frame), vbox);
	
	instruct = gtk_label_new(INSTRUCTIONS);
	level = gtk_progress_bar_new();

	/* Okay and cancel buttons */
	buttonbox = gtk_hbox_new(FALSE, 5);
	button = gtk_button_new_with_label("OK");
	gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(audio_okay), NULL);
	gtk_widget_set_usize(button, 60, 30);
	gtk_box_pack_start(GTK_BOX(buttonbox), button, FALSE, FALSE, 20);

	autoadj = button = gtk_toggle_button_new_with_label("Auto Adjust");
	gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(audio_adjust), NULL);
	gtk_widget_set_usize(button, 80, 30);
	gtk_box_pack_start(GTK_BOX(buttonbox), button, FALSE, FALSE, 20);

	button = gtk_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(audio_cancel), NULL);
	gtk_widget_set_usize(button, 60, 30);
	gtk_box_pack_start(GTK_BOX(buttonbox), button, FALSE, FALSE, 20);
	
	gtk_box_pack_start(GTK_BOX(outerbox), frame, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(outerbox), level, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(outerbox), instruct, FALSE, FALSE, 5);
	gtk_box_pack_end(GTK_BOX(outerbox), buttonbox, FALSE, FALSE, 5);
	gtk_container_set_border_width(GTK_CONTAINER(outerbox), 10);
	
	audiobox = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(audiobox), "Gnophone: Sound settings");
	gtk_container_add(GTK_CONTAINER(audiobox), outerbox);
	gtk_signal_connect(GTK_OBJECT(audiobox), "destroy", GTK_SIGNAL_FUNC(audiodestroy), NULL);

	open_mixerfd();
	return 0;
}

static int show_option_screen(void)
{
	gtk_widget_show_all(audiobox);
	gtk_widget_hide(instruct);
	gtk_widget_hide(level);
	gtk_grab_add(audiobox);
	gtk_adjustment_set_value(GTK_ADJUSTMENT(adj), mainvol>>8);
	gtk_adjustment_set_value(GTK_ADJUSTMENT(adj2), micvol>>8);

	while(audiobox && !gtk_main_iteration_do(TRUE));
	return 0;
}

int show_audio_control(void)
{
	build_option_screen();
	show_option_screen();
	return 0;
}

static void
open_mixerfd(void)
{
	if(!mixer_fd)
		if ( (mixer_fd = open("/dev/mixer", O_RDWR, 0)) < 0)
			perror("/dev/mixer");
	if(ioctl(mixer_fd, SOUND_MIXER_READ_VOLUME, &mainvol) == -1)
			perror("SOUND_MIXER_READ_VOLUME");
	gtk_adjustment_set_value(GTK_ADJUSTMENT(adj), mainvol>>8);

	if(ioctl(mixer_fd, SOUND_MIXER_READ_MIC, &micvol) == -1)
			perror("SOUND_MIXER_READ_MIC");
	gtk_adjustment_set_value(GTK_ADJUSTMENT(adj2), micvol>>8);

	if(ioctl(mixer_fd, SOUND_MIXER_READ_RECLEV, &recvol) == -1)
			perror("SOUND_MIXER_READ_RECLEV");
	gtk_adjustment_set_value(GTK_ADJUSTMENT(adj3), recvol>>8);
}

int set_rec_source(void)
{
	int src = SOUND_MASK_MIC;
	if(!mixer_fd)
		if ( (mixer_fd = open("/dev/mixer", O_RDWR, 0)) < 0)
			perror("/dev/mixer");
	if (mixer_fd)
		if (ioctl(mixer_fd, SOUND_MIXER_WRITE_RECSRC, &src) == -1)
			perror("SOUND_MIXER_RECSRC");
	return 0;
}


static void update_mixer_mainvol(GtkAdjustment *adj)
{
	mainvol = (int)adj->value;

	mainvol = ((mainvol << 8)|mainvol);

	if(mixer_fd)
		if(ioctl(mixer_fd, SOUND_MIXER_WRITE_VOLUME, &mainvol) == -1)
			perror("SOUND_MIXER_WRITE_VOLUME");
}

static void update_mixer_micvol(GtkAdjustment *adj)
{
	micvol = (int)adj->value;

	micvol = ((micvol << 8)|micvol);

	if(mixer_fd)
		if(ioctl(mixer_fd, SOUND_MIXER_WRITE_MIC, &micvol) == -1)
			perror("SOUND_MIXER_WRITE_MIC");
}

static void update_mixer_recvol(GtkAdjustment *adj)
{
	recvol = (int)adj->value;

	recvol = ((recvol << 8)|recvol);

	if(mixer_fd)
		if(ioctl(mixer_fd, SOUND_MIXER_WRITE_RECLEV, &recvol) == -1)
			perror("SOUND_MIXER_WRITE_RECLEV");
}

